/** *************************************************************************
 **
 ** Copyright (C) 2010 Heinz van Saanen
 **
 ** This file is part of the interactive editor CLedit.
 **
 ** CLedit may be used under the terms of the GNU General Public License
 ** version 3 or later as published by the Free Software Foundation and
 ** appearing in the file LICENSE.GPL included in this package.
 **
 ** CLedit is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
 ** WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 **
****************************************************************************/


/** Includes **/
#include <QtGui>
#include <iostream>
#include "high.h"


/** Where was CLedit stared from? **/
extern QString workDir;


/** Parenthesis extra stuff **/
QVector<ParenthesisInfo *> TextBlockData::parentheses() {
	return m_parentheses;
}


void TextBlockData::insert( ParenthesisInfo *info ) {
	int i = 0;
	while (
		i < m_parentheses.size() &&
		info->position > m_parentheses.at(i)->position )
		++i;
	m_parentheses.insert( i, info );
}


/** Lisp rules **/
void Highlighter::loadLispRules() {

	// Keyword/symbol format
	m_keywordFormat.setForeground( Qt::red );
	//m_keywordFormat.setFontWeight( QFont::Bold );

	QFile file( workDir + "/rules/lisp.txt" );
	QTextStream in( &file );
	if( !file.exists() || !file.open(QFile::ReadOnly | QFile::Text)) {
		std::cerr << "ERROR: missing rules for Lisp" << std::endl;
		return;
	}

	// Handle symbols extra
	m_keywordPatterns
		<< "\\+ "
		<< "\\- "
		<< "\\/ "
		<< "\\* ";

	// Read keyword rules in
	QString line;
	do {
		line = in.readLine();
		if ( !line.startsWith("#") && !line.isEmpty() ) {
			m_keywordPatterns << ("\\b" + line + "\\b");
			wordList << line;
		}
	} while ( !line.isNull() );
	file.close();
	wordList.removeDuplicates();
	qSort( wordList.begin(), wordList.end() );

	// Apply keyword rules
	foreach ( const QString &pattern, m_keywordPatterns ) {
		m_rule.pattern = QRegExp( pattern );
		m_rule.format = m_keywordFormat;
		m_highlightingRules.append( m_rule );
	}

	// Comments
	m_commentFormat.setForeground( Qt::darkBlue );
	m_commentFormat.setFontItalic( true );
	m_rule.pattern = QRegExp( ";[^\n]*" );
	m_rule.format = m_commentFormat;
	m_highlightingRules.append( m_rule );

	// Strings
	m_quotationFormat.setForeground( Qt::darkGreen );
	m_rule.pattern = QRegExp( "\".*\"" );
	m_rule.format = m_quotationFormat;
	m_highlightingRules.append( m_rule );
}


/** Constructor **/
Highlighter::Highlighter( QTextDocument *parent ) : QSyntaxHighlighter( parent ) {
}


/** Re-implemented **/
void Highlighter::highlightBlock( const QString &text ) {

	// Lisp rules
	foreach ( const HighlightingRule &m_rule, m_highlightingRules ) {
		QRegExp expression( m_rule.pattern );
		int index = expression.indexIn( text );
		while (index >= 0) {
			int length = expression.matchedLength();
			setFormat( index, length, m_rule.format );
			index = expression.indexIn( text, index + length );
		}
	}

	// Parenthesis
	TextBlockData *data = new TextBlockData;

	int leftPos = text.indexOf( '(' );
	while ( leftPos != -1 ) {
		ParenthesisInfo *info = new ParenthesisInfo;
		info->character = '(';
		info->position = leftPos;

		data->insert( info );
		leftPos = text.indexOf( '(', leftPos+1 );
	}

	int rightPos = text.indexOf(')');
	while ( rightPos != -1 ) {
		ParenthesisInfo *info = new ParenthesisInfo;
		info->character = ')';
		info->position = rightPos;

		data->insert( info );
		rightPos = text.indexOf( ')', rightPos+1 );
	}

	setCurrentBlockUserData(data);
}
