/** *************************************************************************
 **
 ** Copyright (C) 2010 Heinz van Saanen
 **
 ** This file is part of the interactive editor CLedit.
 **
 ** CLedit may be used under the terms of the GNU General Public License
 ** version 3 or later as published by the Free Software Foundation and
 ** appearing in the file LICENSE.GPL included in this package.
 **
 ** CLedit is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
 ** WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 **
****************************************************************************/


/** Includes **/
#include <QtGui>
#include <iostream>
#include "mainwindow.h"


/** Actual CLedit version **/
#define VERSION "0.3.1"
#define PROGNAME ( (QString)"CLedit" )


/** GUI-specific defines **/
#define FONT "Monospace"           // Font used troughout CLedit. Advised a sans serif/monospaced one
#define FONTSIZE 10                // Font size troughout CLedit
#define ICONSIZE QSize( 20, 20 )   // Icon dimensions for toolbar
#define TABWIDTH 24                // Tabulator width in pixel ( 8 per character )


/** (Lispworks) HyperSpec settings. Set paths according to your needs **/
#define USE_LOCAL_HS   0   // Use remote file or local HyperSpec file?
#define HS_LOCAL_FILE  ( QString("hyperspec/Front/index_tx.htm") )
#define HS_WEB_FILE    ( QString("www.lispworks.com/documentation/HyperSpec/Front/Contents.htm") )


/** Where was CLedit stared from? **/
extern QString workDir;


/** Constructor **/
MainWindow::MainWindow( QWidget *parent, Qt::WindowFlags f ) : QMainWindow( parent, f ) {

	// Basic setups
	setupUi( this );
	setWindowTitle( PROGNAME );
	m_historyCnt = 0;
	m_historyCur = 0;
	textEdit->setTabStopWidth( TABWIDTH );

	// Editors
	m_messageEdit = new QPlainTextEdit();
	m_messageEdit->setWindowTitle( tr("Lisp messages") );
	m_messageEdit->setReadOnly( true );
	m_messageEdit->show();

	m_historyEdit = new QPlainTextEdit();
	m_historyEdit->setWindowTitle( tr("REPL history") );
	m_historyEdit->setReadOnly( true );
	m_historyEdit->show();

	m_helpEdit = new QTextBrowser();

	// Load stored settings
	readSettings();

	// Highlighter and help auto-completer
	m_highlighter = new Highlighter( textEdit->document() );
	m_highlighter->loadLispRules();
	m_completer = new QCompleter( m_highlighter->wordList, this );
	helpEdit->setCompleter( m_completer );

	// Application font
	QFont appFnt( FONT, FONTSIZE );
	QApplication::setFont( appFnt );

	// Generate process instance
	m_process = new Process( this );
	m_process->setCommand( m_cmdStr );

	// Setup actions
	setupActions();
	actionExecute_selection->setEnabled( false );
	actionExecute_all->setEnabled( false );
	actionHistory_up->setEnabled( false );
	actionHistory_down->setEnabled( false );
	actionStart->setEnabled( true );
	actionKill->setEnabled( false );

	// Process related connections
	connect( actionStart, SIGNAL( triggered() ), m_process, SLOT( startSlot() ));
	connect( actionKill, SIGNAL( triggered() ), m_process, SLOT( killSlot() ));
	connect( m_process, SIGNAL( messageOut( QString& ) ), this, SLOT( messageOutSlot( QString& ) ));
	connect( m_process, SIGNAL( procState( bool ) ), this, SLOT( procStateSlot( bool ) ));
	connect( actionExecutable, SIGNAL( triggered() ), this, SLOT( executableSlot() ));

	// Menu "Editors"
	connect( actionMessages, SIGNAL( triggered() ), this, SLOT( messagesSlot() ));
	connect( actionHistory, SIGNAL( triggered() ), this, SLOT( historySlot() ));
	connect( actionHistory, SIGNAL( triggered() ), this, SLOT( historySlot() ));
	connect( actionShow_all, SIGNAL( triggered() ), this, SLOT( show_AllSlot() ));

	// Parenthesis matcher
	connect( textEdit, SIGNAL( cursorPositionChanged() ), this, SLOT( matchPar() ));

	// Test if clipboard has content and can paste into document initially
	actionPaste->setEnabled( textEdit->canPaste() );
}


/** Set up actions **/
void MainWindow::setupActions() {

	// Use some standard Icons, so use a QStyle for this
	QStyle *style = QApplication::style();

	// Menu "File"
	actionNew->setIcon( style->standardIcon( QStyle::SP_FileIcon ));
	actionOpen->setIcon( style->standardIcon( QStyle::SP_DirOpenIcon ));
	actionSave_as->setIcon( style->standardIcon( QStyle::SP_DialogSaveButton ));
	actionExit->setIcon( style->standardIcon( QStyle::SP_TitleBarCloseButton ));

	// Menu "Edit"
	actionUndo->setIcon( style->standardIcon( QStyle::SP_ArrowBack ));
	actionRedo->setIcon( style->standardIcon( QStyle::SP_ArrowForward ));
	actionHistory_up->setIcon( style->standardIcon( QStyle::SP_ArrowUp ));
	actionHistory_down->setIcon( style->standardIcon( QStyle::SP_ArrowDown ));

	// Menu "Lisp". Start/kill/determine process here
	actionStart->setIcon( style->standardIcon( QStyle::SP_DialogApplyButton ));
	actionKill->setIcon( style->standardIcon( QStyle::SP_DialogCancelButton ));

	// Menu "Help"
	actionAbout->setIcon( style->standardIcon( QStyle::SP_MessageBoxInformation ));
	actionAbout_Qt->setIcon( style->standardIcon( QStyle::SP_MessageBoxInformation ));

	// Add actions to context in menuEdit
	foreach( QAction *action, menuEdit->actions() )
		textEdit->addAction( action );

	// Menu "File"
	connect( actionNew, SIGNAL( triggered() ), this, SLOT( newSlot() ));
	connect( actionOpen, SIGNAL( triggered() ), this, SLOT( openSlot() ));
	connect( actionSave_as, SIGNAL( triggered() ), this, SLOT( save_asSlot() ));
	connect( actionPrint, SIGNAL( triggered() ), this, SLOT( printSlot() ));
	connect( actionExit, SIGNAL( triggered() ), this, SLOT( exitSlot() ));

	// Menu "Edit"
	connect( actionExecute_selection, SIGNAL( triggered() ), this, SLOT( execute_selectionSlot() ));
	connect( actionExecute_all, SIGNAL( triggered() ), this, SLOT( execute_allSlot() ));
	connect( actionHistory_up, SIGNAL( triggered() ), this, SLOT( history_upSlot() ));
	connect( actionHistory_down, SIGNAL( triggered() ), this, SLOT( history_downSlot() ));

	// Menu "Help"
	connect( actionHelp, SIGNAL( triggered() ), this, SLOT( helpSlot() ));
	connect( actionHyperSpec, SIGNAL( triggered() ), this, SLOT( hyperspecSlot() ));
	connect( actionAbout, SIGNAL( triggered() ), this, SLOT( aboutSlot() ));
	connect( actionAbout_Qt, SIGNAL( triggered() ), qApp, SLOT( aboutQt() ));

	connect( textEdit->document(), SIGNAL( contentsChanged() ), this, SLOT( contentsChangedSlot() ));
	connect( lineEdit, SIGNAL( returnPressed() ), this, SLOT( lineSlot() ));
	connect( helpEdit, SIGNAL( returnPressed() ), this, SLOT( lispHelpSlot() ));
	connect( m_completer, SIGNAL( activated( const QString& )), this, SLOT( completerSlot( const QString& ) ));

	// Toolbars; re-use menu actions
	QToolBar *fileBar;
	fileBar = addToolBar( tr("File") );
	fileBar->setIconSize( ICONSIZE );
	fileBar->addAction( actionNew );
	fileBar->addAction( actionOpen );
	fileBar->addAction( actionSave_as );

	QToolBar *lispBar;
	lispBar = addToolBar( tr("Execute") );
	lispBar->setIconSize( ICONSIZE );
	lispBar->addAction( actionExecute_selection );
	lispBar->addAction( actionExecute_all );

	QToolBar *histoBar;
	histoBar = addToolBar( tr("History") );
	histoBar->setIconSize( ICONSIZE );
	histoBar->addAction( actionHistory_up );
	histoBar->addAction( actionHistory_down );

	QToolBar *exeBar;
	exeBar = addToolBar( tr("Executable") );
	exeBar->setIconSize( ICONSIZE );
	exeBar->addAction( actionStart );
	exeBar->addAction( actionKill );
	exeBar->addAction( actionExecutable );

	// Finally add non-standard or still missing icons
	actionPrint->setIcon( QIcon(":/images/print.png") );
	actionExecute_selection->setIcon( QIcon(":/images/exesel.png") );
	actionExecute_all->setIcon( QIcon(":/images/exeall.png") );
	actionMessages->setIcon( QIcon(":/images/editor.png") );
	actionHistory->setIcon( QIcon(":/images/editor.png") );
	actionShow_all->setIcon( style->standardIcon( QStyle::SP_BrowserReload ));
	actionExecutable->setIcon( QIcon(":/images/exe.png") );
	actionHelp->setIcon( QIcon(":/images/help.png") );
	actionHyperSpec->setIcon( QIcon(":/images/lispworks.png") );
}


/** Menu "File" slots **/
void MainWindow::newSlot() {
	textEdit->clear();
	setWindowTitle( PROGNAME );
}

void MainWindow::openSlot() {
	QString fileName;

	// File-dialog
	if ( !m_fileName.isEmpty() )
		fileName = QFileDialog::getOpenFileName(
						this,
						tr("Open file"),
						m_fileName,
						"Images (*.lisp *.lsp *.sh)");
	else
		fileName = QFileDialog::getOpenFileName(
						this,
						tr("Open file"),
						workDir,
						"Images (*.lisp *.lsp. *.sh)");

	// Some sanity-checks
	if ( fileName.isEmpty() )
		return;

	QFile file(fileName);
	if ( !file.exists() || !file.open(QFile::ReadOnly | QFile::Text) )
		return;

	// And load finally
	textEdit->clear();
	textEdit->setPlainText( file.readAll() );
	file.close();
	setWindowTitle( PROGNAME + " - " + QFileInfo(fileName).fileName() );
	m_fileName = fileName;
}

void MainWindow::save_asSlot() {
	QString fileName;

	// File-dialog
	if ( !m_fileName.isEmpty() )
		fileName = QFileDialog::getSaveFileName(
						this,
						tr("Save file" ),
						m_fileName );
	else
		fileName = QFileDialog::getSaveFileName(
						this,
						tr("Save file" ),
						workDir );

	// Some sanity-checks
	if ( fileName.isEmpty() )
		return;

	QFile file( fileName );
	if ( !file.open(QFile::WriteOnly | QFile::Text) ) {
		QMessageBox::critical(
			this,
			tr("Save"),
			tr("Error saving file:") + " " + fileName );
		return;
	}

	// And save finally
	QTextStream out( &file );
	out << textEdit->toPlainText();
	file.close();
	setWindowTitle( PROGNAME + " - " + QFileInfo(fileName).fileName() );
	m_fileName = fileName;
}

void MainWindow::printSlot() {
	QPrinter printer;
	QPrintDialog printDialog( &printer, this );
	if ( printDialog.exec() == QDialog::Accepted )
		textEdit->print( &printer );
}

void MainWindow::exitSlot() {
	writeSettings();
	m_messageEdit->close();
	m_historyEdit->close();
	m_helpEdit->close();
	m_process->killSlot();
	close();
}


/** Menu "Edit" & toolbar slots **/
void MainWindow::execute_selectionSlot() {

	// Get selected text
	textEdit->copy();
	QString selection;
	QClipboard *clip = QApplication::clipboard();
	selection = clip->text();

	// Care for exact one \n at end of input
	while ( selection.endsWith("\n") )
		selection.resize( selection.size()-1 );
	selection += "\n";

	// And go
	m_process->write( selection );
}

void MainWindow::execute_allSlot() {

	// Select whole text
	textEdit->selectAll();
	textEdit->copy();
	QString selection;
	QClipboard *clip = QApplication::clipboard();
	selection = clip->text();

	// Care for exact one \n at end of input
	while ( selection.endsWith("\n") )
		selection.resize( selection.size()-1 );
	selection += "\n";

	// And go
	m_process->write( selection );
}

// Re-implement key-event for line-editor history
void MainWindow::keyPressEvent(QKeyEvent *event) {
	if ( lineEdit->hasFocus() ) {
		if ( event->key()==Qt::Key_Up )
			history_upSlot();
		if ( event->key()==Qt::Key_Down )
			history_downSlot();
	}
}

// Handle history up/down
void MainWindow::history_upSlot() {
	if ( m_historyCur>0 ) {
		m_historyCur--;
		lineEdit->setText( m_historyList.at( m_historyCur ) );
		lineEdit->setFocus();
	}
	if ( m_historyCur<=0 )
		actionHistory_up->setEnabled(false);
	actionHistory_down->setEnabled(true);
}

void MainWindow::history_downSlot() {
	if ( m_historyCur<m_historyList.size()-1 ) {
		m_historyCur++;
		lineEdit->setText( m_historyList.at( m_historyCur ) );
		actionHistory_up->setEnabled(true);
		lineEdit->setFocus();
	}
	if ( m_historyCur>=m_historyList.size()-1 )
		actionHistory_down->setEnabled(false);
}


/** Menu "Editors". Show/bring up message and history editors **/
void MainWindow::messagesSlot() {
	if ( !m_messageEdit->isVisible() )
		m_messageEdit->show();
	else
		m_messageEdit->showNormal();
	m_messageEdit->raise();
	m_messageEdit->activateWindow();
}

void MainWindow::historySlot() {
	if ( !m_historyEdit->isVisible() )
		m_historyEdit->show();
	else
		m_historyEdit->showNormal();
	m_historyEdit->raise();
	m_historyEdit->activateWindow();
}

void MainWindow::show_AllSlot() {
	messagesSlot();
	historySlot();
}


/** Menu "Lisp"/process slots **/
void MainWindow::messageOutSlot( QString &mess ) {
	m_messageEdit->appendPlainText( mess );
}

void MainWindow::executableSlot() {
	QLabel *textLabel = new QLabel;
	bool ok;
	QString inp = QInputDialog::getText(
						this,
						"Lisp",
						tr("Set new Lisp executable:"),
						QLineEdit::Normal,
						m_cmdStr,
						&ok );
    if ( ok && !inp.isEmpty() ) {
		textLabel->setText( inp );
		m_cmdStr = inp;
		if ( m_process->m_active )
			m_process->killSlot();
		m_process->setCommand( m_cmdStr );
	}
}

void MainWindow::procStateSlot ( bool b ) {
	actionStart->setEnabled( !b );
	actionKill->setEnabled( b );
	actionExecute_selection->setEnabled( b );
	actionExecute_all->setEnabled( b );
}


/** Menu "Help" slots **/
void MainWindow::helpSlot() {   // Use internal browser as default
	const QString hlpFile = workDir + "/doc/cledit.html";
	QFile file( hlpFile );

	if ( !file.exists() ) {
		QMessageBox::warning(
			this,
			"CLedit",
			tr("Cannot find help file") );
		return;
	}

	#if ( defined(WINDOWS) || defined(WIN32) || defined(WIN64) )
	QUrl url(QString( "file:///" ) + hlpFile );
	#else
	QUrl url(QString( "file://" ) + hlpFile );
	#endif

	m_helpEdit->setWindowTitle( (QString)"CLedit " + tr("help") );
	m_helpEdit->setSource( url );
	m_helpEdit->show();
}

void MainWindow::hyperspecSlot() { // Use system browser as default 
	QString hyperspecFile;
	QFile file;
	QUrl url;

	// Use local HyperSpec file?
	if ( USE_LOCAL_HS ) {
		hyperspecFile = workDir + "/" + HS_LOCAL_FILE;
		file.setFileName ( hyperspecFile );

		if ( !file.exists() ) {
			QMessageBox::warning(
				this,
				"CLedit",
				tr("Cannot find local HyperSpec file") );
			return;
		}

		#if ( defined(WINDOWS) || defined(WIN32) || defined(WIN64) )
		url.setUrl( (QString("file:///")) + hyperspecFile );
		#else
		url.setUrl( (QString("file://")) + hyperspecFile );
		#endif
	}

	// Use remote HyperSpec file?
	if ( !USE_LOCAL_HS ) {
		hyperspecFile = HS_WEB_FILE;
		#if ( defined(WINDOWS) || defined(WIN32) || defined(WIN64) )
		url.setUrl( (QString("http:///")) + hyperspecFile );
		#else
		url.setUrl( (QString("http://")) + hyperspecFile );
		#endif
	}

	// And go. For remote files you must be online of course:-)
	QDesktopServices::openUrl( url );
}

void MainWindow::aboutSlot() {

	QMessageBox::about(
		this,
		tr("About") + " CLedit",

		QString("<b>") + "CLedit" + ("</b> ") +
        QString("is an interactive Lisp editor.") +
		QString("<p>Version ") + VERSION + QString("</p>") +
        QString(
			"Contact address: "
			"<a href=\"mailto:zhu3d@aon.at?subject=Zhu3D\""
			">zhu3d@aon.at</a><BR>"

	        "License GPL v3 or later: "
			"<a href=\"http://www.gnu.org/licenses/gpl-3.0-standalone.html\""
			">GPL v3</a></p>") +

		QString("<p>(c) 2010 Heinz van Saanen</p>") );
}


/** History editor **/
void MainWindow::lineSlot() {
	QString inp = lineEdit->text();
	QString sn;

	// Care for exact one \n at end of input
	while ( inp.endsWith("\n") )
		inp.resize( inp.size()-1 );

	// Add missing closing parenthesis automatically
	int opCnt=0, clCnt=0;
	for (int i=0; i<inp.size(); i++)
		if ( inp.at(i)=='(' ) opCnt++;
		else if ( inp.at(i)==')' ) clCnt++;
	if ( opCnt>clCnt )
		for (int i=0; i<opCnt-clCnt; i++)
			inp += ")";
	lineEdit->setText( inp );
	inp += "\n";
	
	// Check blind recursive calls
	if ( !inp.isEmpty() ) {
		m_process->write( inp );
		inp.resize( inp.size()-1 );
		m_historyList.append( inp );
		m_historyEdit->appendPlainText( "CL[" + sn.setNum( m_historyCnt ) + "]>" + inp );
		m_historyCur = ++m_historyCnt;
		actionHistory_down->setEnabled( false );
	}

	// Update toolbar icon
	if ( m_historyCur>=1 )
		actionHistory_up->setEnabled( true );

	lineEdit->clear();
}


/** Lisp command help **/
void MainWindow::completerSlot( const QString &text ) {
	QString dummy = text;
	lispHelpSlot();
}

void MainWindow::lispHelpSlot() {
	QString inp = helpEdit->text();
	QString sn;

	// Care for exact one \n at end of input
	while ( inp.endsWith("\n") )
		inp.resize( inp.size()-1 );
	inp = "(describe \'" + inp + ")";
	inp += "\n";
	
	// Check blind recursive calls
	if ( !inp.isEmpty() ) {
		m_process->write( inp );
		inp.resize( inp.size()-1 );
		m_historyList.append( inp );
		m_historyEdit->appendPlainText( "CL[" + sn.setNum( m_historyCnt ) + "]>" + inp );
		m_historyCur = ++m_historyCnt;
		actionHistory_down->setEnabled( false );
	}

	// Update toolbar icon
	if ( m_historyCur>=1 )
		actionHistory_up->setEnabled( true );

	helpEdit->clear();
}


/** Just a little helper for inital paste **/
void MainWindow::contentsChangedSlot() {
	actionPaste->setEnabled( textEdit->canPaste() );
}


/** *************************************************************************
 ** The cool and basic idea for the parenthesis matcher was published in
 ** Qt Quarterly issue 31/2009 to be found at: http://doc.trolltech.com/qq/
 **
 ** Most of the code was taken from their GPLed example. Major modifications
 ** had to be made in the matchRightParenthesis function. The original was
 ** buggy and worked for most simple and trivial cases only.
****************************************************************************/

/** Parenthesis matcher **/
void MainWindow::matchPar() {

	QList<QTextEdit::ExtraSelection> selections;
	textEdit->setExtraSelections(selections);
	QTextBlock textBlock = textEdit->textCursor().block();
	TextBlockData *data = static_cast<TextBlockData *>( textBlock.userData() );

	if ( data ) {
		QVector<ParenthesisInfo *> infos = data->parentheses();
		int pos = textEdit->textCursor().block().position();

		for ( int i=0; i<infos.size(); ++i ) {
			ParenthesisInfo *info = infos.at(i);
			int curPos = textEdit->textCursor().position() - textBlock.position();

			// Clicked on a left parenthesis?
			if ( info->position == curPos-1 && info->character == '(' ) {
				if ( matchLeftPar( textBlock, i+1, 0 ) )
					createParSelection( pos + info->position );
			}

			// Clicked on a right parenthesis?
			if ( info->position == curPos-1 && info->character == ')' ) {
				if ( matchRightPar( textBlock, i-1, 0 ) )
					createParSelection( pos + info->position );
			}
		}
	}
}

/** Test left parenthesis match **/
bool MainWindow::matchLeftPar(
			QTextBlock currentBlock, int index, int numLeftPar ) {

	TextBlockData *data = static_cast<TextBlockData *>( currentBlock.userData() );
	QVector<ParenthesisInfo *> infos = data->parentheses();
	int docPos = currentBlock.position();

	// Match in same line?
	for ( ; index<infos.size(); ++index ) {
		ParenthesisInfo *info = infos.at(index);

		if ( info->character == '(' ) {
			++numLeftPar;
			continue;
		}

		if ( info->character == ')' && numLeftPar == 0 ) {
			createParSelection( docPos + info->position );
			return true;
		} else
			--numLeftPar;
    }

	// No match yet? Then try next block
	currentBlock = currentBlock.next();
	if ( currentBlock.isValid() )
		return matchLeftPar( currentBlock, 0, numLeftPar );

	// No match at all
    return false;
}

/** Test right parenthesis match **/
bool MainWindow::matchRightPar(
			QTextBlock currentBlock, int index, int numRightPar) {

	TextBlockData *data = static_cast<TextBlockData *>( currentBlock.userData() );
	QVector<ParenthesisInfo *> infos = data->parentheses();
	int docPos = currentBlock.position();

	// Match in same line?
	for (int j=index; j>=0; --j ) {
		ParenthesisInfo *info = infos.at(j);

		if ( info->character == ')' ) {
			++numRightPar;
			continue;
		}

		if ( info->character == '(' && numRightPar == 0 ) {
			createParSelection( docPos + info->position );
			return true;
		} else
			--numRightPar;
	}

	// No match yet? Then try previous block
	currentBlock = currentBlock.previous();
	if ( currentBlock.isValid() ) {

		// Recalculate correct index first
		TextBlockData *data = static_cast<TextBlockData *>( currentBlock.userData() );
		QVector<ParenthesisInfo *> infos = data->parentheses();

		return matchRightPar( currentBlock, infos.size()-1, numRightPar );
	}

	// No match at all
	return false;
}

/** Set parenthesis highlighter at pos **/
void MainWindow::createParSelection( int pos ) {

	QList<QTextEdit::ExtraSelection> selections = textEdit->extraSelections();
	QTextEdit::ExtraSelection selection;
	QTextCharFormat format = selection.format;
	format.setBackground( QColor(0, 20, 200, 40) );
	selection.format = format;
    
	QTextCursor cursor = textEdit->textCursor();
	cursor.setPosition( pos );
	cursor.movePosition( QTextCursor::NextCharacter, QTextCursor::KeepAnchor );

	selection.cursor = cursor;
	selections.append( selection );
	textEdit->setExtraSelections( selections );
}


/** *************************************************************************
 ** Save/reload user settings
****************************************************************************/
void MainWindow::readSettings() {
	QSettings settings("CLedit", PROGNAME);

	// Read window positions
	QPoint mainPos = settings.value( "MainWinPos", QPoint(0, 0) ).toPoint();
	QPoint messPos = settings.value( "MessWinPos" ).toPoint();
	QPoint histPos = settings.value( "HistWinPos" ).toPoint();
	QPoint helpPos = settings.value( "HelpWinPos" ).toPoint();

	// Read window sizes
	QSize mainSize = settings.value( "MainWinSize", QSize(0, 0) ).toSize();
	QSize messSize = settings.value( "MessWinSize" ).toSize();
	QSize histSize = settings.value( "HistWinSize" ).toSize();
	QSize helpSize = settings.value( "HelpWinSize" ).toSize();

	// Read exe/process
	QString lispExe = settings.value( "LispExe" ).toString();

	// Settings yet saved at least once? Reload them...
	if( mainSize.width() != 0 ) {
		move( mainPos );
		m_messageEdit->move( messPos );
		m_historyEdit->move( histPos );
		m_helpEdit->move( helpPos );

		resize( mainSize );
		m_messageEdit->resize( messSize );
		m_historyEdit->resize( histSize );
		m_helpEdit->resize( helpSize );

		m_cmdStr = lispExe;
	}
	// Else 1.start ever! So use plausible values for smaller screens
	else {
		move( QPoint(0, 0) );
		m_messageEdit->move( QPoint(420, 0) );
		m_historyEdit->move( QPoint(840, 0) );
		m_helpEdit->move( QPoint(200, 50) );

		resize( QSize(400, 700) );
		m_messageEdit->resize( QSize(400, 700) );
		m_historyEdit->resize( QSize(300, 700) );
		m_helpEdit->resize( QSize(600, 700) );

		m_cmdStr = "sbcl";
	}
}

void MainWindow::writeSettings() {
	QSettings settings( "CLedit", PROGNAME );

	settings.setValue( "MainWinPos", pos() );
	settings.setValue( "MessWinPos", m_messageEdit->pos() );
	settings.setValue( "HistWinPos", m_historyEdit->pos() );
	settings.setValue( "HelpWinPos", m_helpEdit->pos() );

	settings.setValue( "MainWinSize", size() );
	settings.setValue( "MessWinSize", m_messageEdit->size() );
	settings.setValue( "HistWinSize", m_historyEdit->size() );
	settings.setValue( "HelpWinSize", m_helpEdit->size() );

	settings.setValue( "LispExe", m_cmdStr );
}
