/** *************************************************************************
 **
 ** Copyright (C) 2010 Heinz van Saanen
 **
 ** This file is part of the interactive editor CLedit.
 **
 ** CLedit may be used under the terms of the GNU General Public License
 ** version 3 or later as published by the Free Software Foundation and
 ** appearing in the file LICENSE.GPL included in this package.
 **
 ** CLedit is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
 ** WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 **
****************************************************************************/


/** Includes **/
#include <QDateTime>
#include <iostream>
#include "process.h"


/** Constructor **/
Process::Process( QObject *parent ) : QProcess( parent ) {

	m_active = false;
	m_process.setProcessChannelMode(QProcess::MergedChannels);

	connect(&m_process, SIGNAL(finished(int, QProcess::ExitStatus)), this, SLOT(exitSlot(int, QProcess::ExitStatus)));
	connect(&m_process, SIGNAL(readyReadStandardOutput()), this, SLOT(readSlot()));
}


/** Determine executable to start **/
void Process::setCommand( QString &cmd ) {
	m_cmdStr = cmd;
}


/** Start process **/
void Process::startSlot() {
	if ( m_active ) {
		m_message = tr("PROCESS: already running\n");
		emit messageOut( m_message );
		return;
	}

	m_process.start( m_cmdStr );
	if ( !m_process.waitForStarted() ) {
		m_message = tr("PROCESS: failed to start\n");
		m_active = false;
		emit messageOut( m_message );
		emit procState ( m_active );
		return;
	}

	m_active = true;
	m_message = tr("Session started: %1\n").arg( QDateTime::currentDateTime().toString() );
	emit messageOut( m_message );
	emit procState ( m_active );
}


/** Read data from process stdout and stdcerr **/
void Process::readSlot() {
	char buf;
	QString in;
	while ( m_process.bytesAvailable() ) {
		m_process.getChar(&buf);
		in += buf;
	}
	if ( in.endsWith('\n') )
		in.resize(in.size()-1);
	m_message = in;
	emit messageOut( m_message );
}


/** Write data **/
void Process::write(QString &cmd) {
	m_process.write(cmd.toLocal8Bit().data());
	if (cmd == "(quit)\n") {
		m_message = tr("\nSession ended : %1\n").arg( QDateTime::currentDateTime().toString() );
		emit messageOut( m_message );
	}
}


/** Kill process **/
void Process::killSlot() {
	if ( !m_active ) {
		m_message = tr("PROCESS: nothing to terminate\n");
		emit messageOut( m_message );
		emit procState ( m_active );
		return;
	}
	m_process.close();
	m_active = false;
	m_message = tr("\nSession ended : %1\n").arg( QDateTime::currentDateTime().toString() );
	emit messageOut( m_message );
	emit procState ( m_active );
}


/** Some exit error handling **/
void Process::exitSlot( int exitCode, QProcess::ExitStatus exitStatus ) {

	if ( exitStatus != QProcess::NormalExit )
		std::cerr << "PROCESS: terminated or crashed" << std::endl;

	std::cerr << ( exitCode ? "PROCESS: exit failed" : "PROCESS: exit ok" ) << std::endl;
}
